#!/usr/bin/python
from umpire.deploy import DeploymentModule

import unittest
import os
import shutil
import sys
import tempfile

deployment_json = """
[
    {
        "url":"s3://umpire-test/",
        "items":[
            {
                "name":"test",
                "version":"test_zip",
                "platform":"test",
                "destination":"$tc3/testA"
            },
            {
                "name":"test",
                "version":"test_tgz",
                "platform":"test",
                "destination":"$tc3/testB"
            }
        ]
    }
]
"""

class TC3(unittest.TestCase):
    """
    1. Full deploy from install
    - Run deployment file
    """

    #Currently only on linux
    tempdir = os.path.join(tempfile.gettempdir(), "tc3")
    cache_root = os.path.join(tempdir,"./tc3_cache")
    deployment_file = os.path.join(tempdir,"./tc3_deploy")
    environment_variables = { "tc3" : os.path.join(tempdir,"deployment") }

    # preparing to test
    def setUp(self):
        print("Setting up environment")
        for k, v in list(self.environment_variables.items()):
            os.environ[k] = v
        print("Creating tmp folder.. " + str(self.tempdir))
        os.mkdir(self.tempdir)
        print("Creating deployment file..")
        with open(self.deployment_file, "w+") as f:
            f.write(deployment_json)
    # ending the test
    def tearDown(self):
        """Remove test folder"""
        print("Removing temporary folder")
        try:
            shutil.rmtree(self.tempdir)
        except:
            print("Warning, cannot clean up tempdir.")

    # test routine A
    def test(self):
        print("Running first deploy...")
        self.deploy()
        self.verify_deploy()
        print("Verified files exist")
        print("Deleting cache directory")
        shutil.rmtree(self.cache_root)
        print("Deleting deployment directory")
        shutil.rmtree(self.environment_variables["tc3"])
        print("Running deploy...")
        self.deploy()
        self.verify_deploy()
        print("Verified links still valid.")

    def deploy(self):
        print("Getting deployment module")
        deploy = DeploymentModule()
        deploy.cache_root = self.cache_root
        deploy.deployment_file = self.deployment_file
        deploy.DEBUG = True
        print("Starting deploy and waiting for finish.")
        deploy.run(None)

    def verify_deploy(self):
        with open(os.path.join(self.tempdir,"./deployment/testA/this_is_test_data")) as f:
            lines = f.readlines()
            self.assertEqual(lines[0],"HI!\n")

        with open(os.path.join(self.tempdir,"./deployment/testB/this_is_test_data")) as f:
            lines = f.readlines()
            self.assertEqual(lines[0],"HI!\n")



from maestro.core import execute
from maestro.aws import s3
from maestro.tools import path

__version__ = "0.5.5"

import sys,os

#Internal Modules
from . import fetch, unpack, deploy, update

from .deploy import HELPTEXT

#Config
from . import config

#Entry point for Umpire
def entry():
    exit_code = Umpire().exit_code
    sys.exit(exit_code)

class Umpire(execute.ModuleExecuter):

    #Defaults for command line flags
    skip_update = True
    debug = False
    deployment_file = None

    #
    # !!!Register all the dependencies of this program here.!!!
    #
    def register_dependencies(self):
        self.register(s3.AsyncS3Downloader())
        self.register(fetch.FetchModule())
        self.register(unpack.UnpackModule())
        self.register(deploy.DeploymentModule())
        self.register(update.UpdateModule())

    def run(self, kwargs):

        for index, item in enumerate(sys.argv):
            if index == 0:
                continue
            if item == "-c" or item == "--clear-cache":
                import shutil
                shutil.rmtree(get_umpire_root())
                sys.exit(0)
            elif item == "-r" or item == "--repair-cache":
                path.purge(config.LOCK_FILENAME, get_umpire_root())
                sys.exit(0)
            elif item == "-h" or item == "--help":
                print(deploy.HELPTEXT)
                sys.exit(0)
            elif item == "--version":
                from subprocess import call
                call(["pip3","show","umpire"])
                sys.exit(0)
            elif item == "-s" or item == "--skip-update":
                self.skip_update = True
            elif item == "-d" or item == "--debug":
                self.debug = True
            else:
                self.deployment_file = item


        self.register_dependencies()
        if not self.skip_update:
            updater = self.getObject("update")
            try:
                #Check if there's an update available
                with open(os.path.join(config.REMOTE_VERSION_FILENAME), "r") as f:
                    version = f.read()
                    rmajor, rminor, rrev = update.parse_version_string(version)
                    major, minor, rev = update.parse_version_string(__version__)
                    print (str(rmajor) + str(rminor) + str(rrev))
                    print (str(major) + str(minor) + str(rev))
                    if (rmajor > major) or ((rminor > minor) and (rmajor == rmajor)) or ((rrev > rev) and (rminor == minor) and (rmajor == rmajor)):
                        if config.autoupdate:
                            print("\n!!!Update (%s) detected!!!\n" % version[:-1])
                            print("Autoupdating...\n")
                            updater.run({"update":None})
                            sys.exit(0)
                        else:
                            print("!!!Update (%s) detected!!!" % version[:-1])
                            print("Please run pip install umpire --upgrade as an Administrator.\n")
                            updater.start()
            except IOError:
                updater.start()



        #Get the instance of our first module
        deployer = self.getObject("deploy")

        #TODO: Fix to use just RUN
        deployer.cache_root = get_umpire_root()

        deployer.deployment_file = self.deployment_file

        deployer.DEBUG = self.debug

        self.exit_code = deployer.run(kwargs)

def get_umpire_root():
    hardcoded_root = "./umpire_tmp"
    try:
        if not os.path.exists(config.default_umpire_root):
            os.makedirs(config.default_umpire_root)
        return config.default_umpire_root
    except Exception as e:
        print("Umpire: Error obtaining default umpire root location, using local directory './umpire_tmp': " + str(e), file=sys.stderr)
        if not os.path.exists(hardcoded_root):
            os.makedirs(hardcoded_root)
        return hardcoded_root



"""deploy.py"""

HELPTEXT = """
                  ----- Umpire -----

Umpire reads a properly formatted JSON deployment file to deploy files.
Examples can be found on GitHub: https://github.com/Signiant/umpire

Usage:  umpire <deployment_file>

Options:
--clear-cache, -c:   Clears the default Umpire cache of all packages
--help, -h:          Displays this help text
--repair-cache, -r:  Removes lock files from the cache
--version:           Displays the current version of Umpire

"""

# Define WindowError if we're on Linux
try:
    WindowsError
except NameError:
    WindowsError = None


import sys, os, json, time, traceback, shutil
from distutils import dir_util
from maestro.core import module
from maestro.tools import path

#Local modules
from . import fetch, cache
# from cache import CacheError

CACHE_ROOT_KEYS = ["c", "with-cache"]
HELP_KEYS = ["h", "help"]

## The following code


## The following code backports an islink solution for windows in python 2.7.x
## It gets assigned to a function called "islink"
def islink_windows(path):
    if os.path.exists(path):
        if os.path.isdir(path):
            FILE_ATTRIBUTE_REPARSE_POINT = 0x0400
            attributes = ctypes.windll.kernel32.GetFileAttributesW(unicode(path))
            return (attributes & FILE_ATTRIBUTE_REPARSE_POINT) > 0
        else:
            command = ['dir', path]
            try:
                with open(os.devnull, 'w') as NULL_FILE:
                    o0 = check_output(command, stderr=NULL_FILE, shell=True)
            except CalledProcessError as e:
                print (e.output)
                return False
            o1 = [s.strip() for s in o0.split('\n')]
            if len(o1) < 6:
                return False
            else:
                return 'SYMLINK' in o1[5]
    else:
        return False

islink = os.path.islink
if os.name == "nt":
    import ctypes
    from subprocess import CalledProcessError, check_output
    islink = islink_windows

class DeploymentError(Exception):
    pass

class DeploymentModule(module.AsyncModule):
    # Required ID of this module
    id = "deploy"

    #Cache Root
    cache_root = None

    #Deployment File
    deployment_file = None

    #Set to true to view tracebacks for exceptions
    DEBUG = False

    def help(self):
        print(self.help_text)
        exit(0)

    def run(self,kwargs):
        try:
            with open(self.deployment_file) as f:
                data = json.load(f)
        except TypeError:
            print(HELPTEXT)
            sys.exit(1)
        except IOError as e:
            if not self.DEBUG:
                print("Unable to locate file: " + self.deployment_file)
            else:
                raise e
            sys.exit(1)

        fetchers = list()

        for repo in data:
            repo_url = repo["url"]
            for item in repo["items"]:

                #Required parameters
                name = item["name"]
                version = item["version"]
                platform = item["platform"]
                destination = os.path.expandvars(item["destination"])

                #Optional parameters
                link = True
                try:
                    link = item["link"]
                except KeyError:
                    #Debugging logging
                    pass
                unpack = True
                try:
                    unpack = item["unpack"]
                except KeyError:
                    #Debugging logging
                    pass
                keep_updated = False
                try:
                    keep_updated = item["keep_updated"]
                except KeyError:
                    pass

                #Configure a Fetch module for each entry
                fetcher = fetch.FetchModule(None)
                fetcher.dependency_name = name
                fetcher.dependency_version = version
                fetcher.dependency_platform = platform
                fetcher.dependency_repo = repo_url
                fetcher.dependency_is_link = link
                fetcher.dependency_unpack = unpack
                fetcher.cache_root = self.cache_root
                fetcher.keep_updated = keep_updated
                fetcher.DEBUG = self.DEBUG
                #TODO: Figure out how to move this out of deploy
                try:
                    cache_dir =  os.path.join(fetcher.cache_root, fetcher.get_cache_name())
                    cache.create_local_cache(cache_dir, repo_url)
                except cache.CacheError:
                    pass #Cache already exists
                fetcher.start()
                fetchers.append((fetcher,destination))

        done_count = 0
        exit_code = 0
        while done_count < len(fetchers):
            done_count = 0
            for fetcher, destination in fetchers:
                if fetcher.status == module.PROCESSED:
                    done_count += 1
                    continue
                if not os.path.exists(destination):
                    try:
                        os.makedirs(destination)
                    except OSError as e:
                        if(self.DEBUG):
                            traceback.print_exc()
                        raise DeploymentError(fetcher.format_entry_name() + ": Error attempting to create destination directories.")

                if fetcher.status == module.DONE and fetcher.exception is not None:
                    if self.DEBUG:
                        print (fetcher.exception.traceback)
                    else:
                        print (fetcher.format_entry_name() + ": ERROR -- " + str(fetcher.exception))
                    exit_code = 1
                    fetcher.status = module.PROCESSED
                if fetcher.status == module.DONE and fetcher.exception is None:
                    files = fetcher.result[0]
                    state = fetcher.result[1]
                    for entry in files:
                        destination_file = os.path.join(destination,os.path.split(entry)[1])

                        # If the file exists, and points to the same target as the entry
                        if (os.path.exists(destination_file) and islink(destination_file) and state == fetch.EntryState.CACHE and os.path.realpath(entry) == os.path.realpath(destination_file)):
                            print (fetcher.format_entry_name() + ": Already deployed.")
                            fetcher.status = module.PROCESSED
                            break

                        # If the file exists, but something changed between the entry and the destination_file
                        elif (os.path.exists(destination_file) and (state == fetch.EntryState.UPDATED or state == fetch.EntryState.CACHE or state == fetch.EntryState.DOWNLOADED)):
                            print (fetcher.format_entry_name() + ": Updating " + destination_file)
                            self.__remove_and_deploy_to_destination__(fetcher, entry, destination_file)

                        # Wasn't in the cache, or updated.
                        else:
                            print (fetcher.format_entry_name() + ": Deploying " + destination_file)
                            self.__remove_and_deploy_to_destination__(fetcher, entry, destination_file)
                        #TODO: Kinda hacky, no significance other than to make it not DONE
                        fetcher.status = module.PROCESSED

                if fetcher.status == module.PROCESSED:
                    done_count += 1
            time.sleep(0.1)
        return exit_code

    def __remove_and_deploy_to_destination__(self, fetcher, entry, destination_file):
        if os.path.exists(destination_file):
            try:
                if os.path.isdir(destination_file) and not islink(destination_file):
                    try:
                        os.rmdir(destination_file)
                    except OSError as e:
                        shutil.rmtree(destination_file)
                else:
                    os.unlink(destination_file)
            except OSError as e:
                if(self.DEBUG):
                    traceback.print_exc()
                    raise DeploymentError(fetcher.format_entry_name() + ": Unable to remove previously deployed file: " + str(destination_file))
        try:
            if fetcher.dependency_is_link:
                path.symlink(entry, destination_file)
            elif os.path.isdir(entry):
                dir_util.copy_tree(entry, destination_file)
            else:
                shutil.copyfile(entry, destination_file)
        except WindowsError as e:
            if(self.DEBUG):
                traceback.print_exc()
                raise DeploymentError(fetcher.format_entry_name() + ": Unable to create symlink. Ensure you are running Umpire as an administrator or otherwise enabled your user to create symlinks. Contact your system administrator if this problem persists.")
        except OSError as e:
            if(self.DEBUG):
                traceback.print_exc()
                raise DeploymentError(fetcher.format_entry_name() + ": Unable to create symlink: " + str(e))




"""unpack.py"""

HELPTEXT = """
     ----- Unpack Module -----

The unpack module unpacks a tar gz or zip file.

Usage: Currently N/A

"""

import sys, os
from maestro.core import module

HELP_KEYS = ["h", "help"]

class UnpackError(Exception):
    pass

class UnpackModule(module.AsyncModule):
    # Required ID of this module
    id = "unpack"
    file_path = None
    destination_path = None
    delete_archive = False

    def help(self):
        print(self.help_text)
        exit(0)

    def untar_gzip(self):
        import tarfile
        with tarfile.open(self.file_path, "r:gz") as f:
            f.extractall(self.destination_path)

    def unzip(self):
        import zipfile
        with zipfile.ZipFile(self.file_path) as zf:
            zf.extractall(self.destination_path)

    def run(self,kwargs):
        if self.file_path.endswith("tar.gz") or self.file_path.endswith("tgz"):
            self.untar_gzip()
        elif self.file_path.endswith(".zip"):
            self.unzip()
        else:
            raise UnpackError("Unable to unpack this type of file: " + os.path.split(self.file_path)[1])

        if self.delete_archive is True:
            os.remove(self.file_path)



"""
repo.py contains code to control the local cache for umpire. It is not a module.
"""
#TODO: Add support for tbz/tar.bz

import os, shutil, time, traceback
import configparser
import maestro.tools.path
import maestro.tools.file
import maestro.core.module
from . import unpack
from maestro.tools.os_tools import check_pid
from . import config

# Cache constants
CONFIG_FILENAME = config.CONFIG_FILENAME
CONFIG_REPO_SECTION_NAME = config.CONFIG_REPO_SECTION_NAME
CONFIG_ENTRY_SECTION_NAME = config.CONFIG_ENTRY_SECTION_NAME
LOCK_FILENAME = config.LOCK_FILENAME
CURRENT_ENTRY_CONFIG_VERSION = config.CURRENT_ENTRY_CONFIG_VERSION
CURRENT_REPO_CONFIG_VERSION = config.CURRENT_REPO_CONFIG_VERSION

def backoff(seconds=5):
    """
    Backoff for a random interval of time up to the amount in seconds provided (default: 5)
    """
    import random
    backoff_time = int(random.random() * seconds * 10)
    time.sleep(backoff_time)

def create_local_cache(local_path, remote_url):
    """
    Creates a cache and returns a LocalCache object. NOT THREAD-SAFE.
    """
    cache_config = os.path.join(local_path, CONFIG_FILENAME)
    if os.path.exists(cache_config):
        raise CacheError("A cache already exists in the specified location: " + local_path)

    try:
        os.makedirs(local_path)
    except OSError as e:
        pass #Directories exist

    ## Start creating config
    config = configparser.SafeConfigParser()
    config.add_section(CONFIG_REPO_SECTION_NAME)

    config.set(CONFIG_REPO_SECTION_NAME, "remote_url", remote_url)
    config.set(CONFIG_REPO_SECTION_NAME, "config_version", CURRENT_REPO_CONFIG_VERSION)

    with open(cache_config, "w+") as f:
        config.write(f)

    return LocalCache(local_path)

def get_lockfile_info(lockfile):
    with open(lockfile, 'r') as lf:
        try:
            lockfile_line = lf.readline()
            lockfile_content = lockfile_line.split("::")
            lockfile_host_id = lockfile_content[0]
            lockfile_pid = int(lockfile_content[1])
            return lockfile_host_id, lockfile_pid
        except (IndexError, ValueError):
            #Lockfile seems like it might be corrupted. We'll back off. The writing process(es) should have detected this, and will remove and backoff
            raise EntryLockError("Lockfile appears corrupted. Try running 'umpire -r' and retrying.")

def read_entry(file_location):
    if not os.path.exists(file_location):
        raise EntryError("Unable to find entry file under: " + file_location)

    #Get parser
    parser = configparser.SafeConfigParser()
    try:#Read
        config = parser.read(file_location)
    except configparser.ParsingError:
        raise EntryError("Error parsing the entry configuration file: " + file_location + " Please contact an administrator for help fixing this problem.")

    #Verify the config has been read
    if config is None:
        raise EntryError("The config file: " + file_location + " appears to be empty. Please verify this is the location of a valid umpire cache entry.")

    #Verify the section we're looking for is there
    if CONFIG_ENTRY_SECTION_NAME not in parser.sections():
        raise EntryError("The config file " + file_location + " appears to be missing the [umpire] section. Please verify this is the location of a valid umpire cache entry, or delete the containing folder and try again.")

    entry = EntryInfo()

    #Set file location
    entry.path = os.path.split(file_location)[0]

    #Attempt to parse the md5
    try:
        entry.md5 = parser.get(CONFIG_ENTRY_SECTION_NAME, "md5")
    except configparser.Error:
        pass #TODO: Future
    if not entry.md5:
        pass #TODO: Future

    #Attempt to parse the entry product name
    try:
        entry.name = parser.get(CONFIG_ENTRY_SECTION_NAME, "name")
    except configparser.Error:
        raise EntryError("Unable to parse name from entry file: " + file_location)
    if not entry.name:
        raise EntryError("Unable to parse name from entry file: " + file_location)

    #Attempt to parse the entry platform name
    try:
        entry.platform = parser.get(CONFIG_ENTRY_SECTION_NAME, "platform")
    except configparser.Error:
        raise EntryError("Unable to parse platform from entry file: " + file_location)
    if not entry.name:
        raise EntryError("Unable to parse platform from entry file: " + file_location)

    #Attempt to parse the entry version
    try:
        entry.version = parser.get(CONFIG_ENTRY_SECTION_NAME, "version")
    except configparser.Error:
        raise EntryError("Unable to parse version from entry file: " + file_location)
    if not entry.name:
        raise EntryError("Unable to parse version from entry file: " + file_location)

    #Attempt to parse the entry config version
    try:
        entry.config_version = parser.get(CONFIG_ENTRY_SECTION_NAME, "config_version")
    except configparser.Error:
        raise EntryError("Unable to parse config_version from entry file: " + file_location)
    if not entry.config_version:
        raise EntryError("Unable to parse config_version from entry file: " + file_location)

    try:
        entry.config_version = parser.get(CONFIG_ENTRY_SECTION_NAME, "md5")
    except configparser.Error:
        entry.md5 = None

    return entry

def write_entry(entry):

        ## Start creating config
        config = configparser.SafeConfigParser()
        config_path = os.path.join(entry.path, CONFIG_FILENAME)
        config.add_section(CONFIG_ENTRY_SECTION_NAME)

        config.set(CONFIG_ENTRY_SECTION_NAME, "name", entry.name)
        config.set(CONFIG_ENTRY_SECTION_NAME, "platform", entry.platform)
        config.set(CONFIG_ENTRY_SECTION_NAME, "version", entry.version)
        config.set(CONFIG_ENTRY_SECTION_NAME, "config_version", CURRENT_ENTRY_CONFIG_VERSION)

        if entry.md5:
            config.set(CONFIG_ENTRY_SECTION_NAME, "md5", entry.md5)

        with open(os.path.join(entry.path, CONFIG_FILENAME), "w+") as f:
            return config.write(f)

class EntryInfo(object):
    config_version = None
    md5 = None
    name = None
    path = None
    platform = None
    version = None

class EntryError(Exception):
    pass

class EntryLockError(Exception):
    pass

class EntryLockTimeoutError(Exception):
    pass

class CacheError(Exception):
    pass

class LocalCache(object):
    local_path = None
    settings_file = None
    remote_url = None
    config_version = None
    host_id = None

    #Lock timeout in seconds
    lock_timeout = None

    #Verifies local existance
    def __init__(self, cache_root, lock_timeout = 1800, host_id="localhost"):
        if not os.path.exists(cache_root):
            raise CacheError("The path '" + str(cache_root) + "' does not contain a valid umpire cache.")

        self.local_path = cache_root
        self.lock_timeout = lock_timeout
        self.host_id = host_id
        self.settings_file = os.path.join(cache_root,CONFIG_FILENAME)

        if not os.path.exists(self.settings_file):
            raise CacheError("The path '" + str(cache_root) + "' does not appear to be a valid umpire cache")
        config = None

        #Get parser
        parser = configparser.SafeConfigParser()
        try:#Read
            config = parser.read(self.settings_file)
        except configparser.ParsingError:
            raise CacheError("Error parsing the cache configuration file. Please contact an administrator for help fixing this problem.")

        #Verify the config has been read
        if config is None:
            raise CacheError("The config file appears to be empty. Please verify this is the location of a valid umpire cache.")

        #Verify the section we're looking for is there
        if CONFIG_REPO_SECTION_NAME not in parser.sections():
            raise CacheError("The config file appears to be missing the [umpire] section. Please verify this is the location of a valid umpire cache.")

        #Attempt to parse the data
        try:
            self.remote_url = parser.get(CONFIG_REPO_SECTION_NAME, "remote_url")
        except configparser.Error:
            raise CacheError("Unable to determine the remote URL of this cache" + self.local_path)

        if not self.remote_url:
            raise CacheError("Unable to determine the remote URL of this cache:" + self.local_path)

        try:
            self.config_version = parser.get(CONFIG_REPO_SECTION_NAME, "config_version")
        except configparser.Error:
            raise CacheError("Unable to determine the version of this cache: " + self.local_path)

        if not self.config_version:
            raise CacheError("Unable to determine the version of this cache: " + self.local_path)

    #Gets EntryInfo object, returns None if it doesn't exist.
    def get(self, platform, name, version):
        cs_entry_file = os.path.join(self.local_path, platform, name, version, ".umpire")
        entry_file = maestro.tools.path.get_case_insensitive_path(path=cs_entry_file)
        if entry_file is None:
            return None
        else:
            return read_entry(entry_file)

    #Locks a local entry
    def lock(self, path, retry_count=5):
        """
        Locks a cache entry in order to perform actions on the cache.
        """
        lockfile = os.path.join(path, LOCK_FILENAME)
        timeout_counter = 0
        while(True):
            #Attempt to make parent directories (we don't care if this w/ OSError, which means the folders likely already exist)
            try:
                os.makedirs(path)
            except OSError as e:
                pass
            #Big ol' try, anything we catch in here we want to retry up to retry_count
            try:
                #See if we're already locked
                if(os.path.exists(lockfile)):
                    lock_id,lock_pid = get_lockfile_info(lockfile)
                    #Determine if the lockfile is from this host
                    if lock_id == self.host_id:
                        #... is it me?
                        if lock_pid == os.getpid():
                            if self.DEBUG:
                                print("WARN: Found a lockfile that apparently belongs to this process... unlocking.")
                            self.unlock(path)
                        #It's not us, but we can wait if the process is still running
                        elif check_pid(lock_pid):
                            if self.DEBUG:
                                print("INFO: Umpire is waiting " + str((int(config.LOCKFILE_TIMEOUT)-int(timeout_counter)))  + " for an entry to unlock.")
                        #Otherwise we're busting open..
                        else:
                            print("WARN: Removing lockfile from previous Umpire run")
                            self.unlock(path, force=True)
                            continue
                    else: #We really don't know what's going on with this entry, we'll wait the timeout at most before forcing an unlock
                        if self.DEBUG:
                            print("INFO: Umpire is waiting " + str((int(config.LOCKFILE_TIMEOUT)-int(timeout_counter)))  + " for an entry to unlock.")

                    if timeout_counter >= config.LOCKFILE_TIMEOUT:
                        raise EntryLockTimeoutError("Timed out trying to unlock lockfile: " + str(lockfile))

                    timeout_counter += 10
                    time.sleep(10)

                #Write lockfile
                with open(lockfile, 'w') as lf:
                    lf.write(str(self.host_id) + "::" + str(os.getpid()))
                    lf.close()

                #Read back lockfile
                lock_id,lock_pid = get_lockfile_info(lockfile)
                if lock_id == self.host_id and lock_pid == os.getpid():
                    return
                else:
                    raise EntryLockError("Expected to have lock, but lockfile does not contain the correct information.")
            except Exception as e:
                if retry_count <= 0:
                    print("ERROR: Unable to unlock cache entry after several attempts: ")
                    raise e
                else:
                    backoff()
                    if self.DEBUG:
                        print("ERROR: Caught the following exception: " + str(e))
                        traceback.print_exc()
                    retry_count -= 1

    def unlock(self, path, force=False):
        lockfile = os.path.join(path, LOCK_FILENAME)
        pid = -1
        #Read back lockfile
        lock_id,lock_pid = get_lockfile_info(lockfile)
        if (lock_id == self.host_id and lock_pid == os.getpid()) or force is True:
            os.remove(lockfile)
        else:
            raise EntryLockError("This process (" + str(os.getpid()) + ") is not the owner (" + str(pid) + ") of the lockfile it's trying to unlock: " + str(lockfile))

    #Puts an archive of files into the cache
    def put(self, archive_path, platform, name, version, unpack_bol=True, force=False, keep_archive = False, keep_original = False, checksum = None):

        if checksum is None:
            #Generate md5
            checksum = maestro.tools.file.md5_checksum(archive_path)

        #Get Archive filename
        archive_filename = os.path.split(archive_path)[1]

        #Get path it would exist in the repo
        full_path = os.path.join(self.local_path, platform, name, version, archive_filename)
        entry_root = os.path.join(self.local_path, platform, name, version)
        case_insensitive_path = maestro.tools.path.get_case_insensitive_path(entry_root)

        #Purge all files in the cache location already that aren't .umplock
        maestro.tools.path.purge("(?!\.umplock).*", case_insensitive_path, match_directories=True)

        #Copy archive to cache location
        shutil.copy(archive_path, case_insensitive_path)

        #Remove original if we want
        if keep_original is False:
            os.remove(archive_path)

        unpacker = unpack.UnpackModule(None)

       #Unpack if necessary
        if unpack_bol is True:
            unpacker.destination_path = case_insensitive_path
            unpacker.file_path = full_path
            unpacker.delete_archive = not keep_archive
            unpacker.start()

        entry = EntryInfo()
        entry.name = name
        entry.version = version
        entry.platform = platform
        entry.path = case_insensitive_path
        entry.md5 = checksum

        #Write the entry
        write_entry(entry)
        if unpack_bol is True:
            while unpacker.status != maestro.core.module.DONE:
                time.sleep(0.2)

            if unpacker.exception is not None:
                pass #Hack for now in case whatever it is isn't an archive
                #raise unpacker.exception

    #Set the repositories remote URL (updates local .umpire file)
    def set_remote(url):
        pass #TODO: Future




from setuptools import setup, find_packages
from distutils.core import setup

from os import path
# this_directory = path.abspath(path.dirname(__file__))
# with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
#     long_description = f.read()

setup(name='Umpire',
      version='0.6.4',
      description='Generic dependency resolver.',
      long_description='',
      long_description_content_type='text/markdown',
      author='Matthew Corner',
      author_email='mcorner@signiant.com',
      url='https://www.signiant.com',
      packages=find_packages(),
      license='MIT',
      install_requires=['MaestroOps>=0.8.8,<0.9'],
      entry_points = {
          'console_scripts': [
              'umpire = umpire.umpire:entry'
              ]
          }
     )
